# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""

@author: Jesus Salgado
@contact: jesusjuansalgado@gmail.com

European Space Astronomy Centre (ESAC)
European Space Agency (ESA)

Created on 15 July 2020
"""
