# Autogenerated by Astropy-affiliated package astroquery's setup.py on 2022-03-23 06:55:41 UTC
import datetime

version = "0.4.6"
githash = "7e4468a2bc25d48da109e182cc967333051f9d8f"


major = 0
minor = 4
bugfix = 6

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2022, 3, 23, 6, 55, 41)
debug = False

astropy_helpers_version = "4.0.1"
